#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Different non-specific tools widely used along the project.
"""

import time
import os
from typing import Optional, List
import logging
from pathlib import Path

from gisaxs_xpcs.read_edf import read_header_from_file

_DEDFAULT_MONETA_DIR = '/home/vladimir/remote_folder/OMBD/'


class MonetaDirNotSetError(ValueError):
    """Please, set the correct path to OMBD folder on moneta server. """
    pass


class WrongOMBDDirError(ValueError):
    """Please, check that the provided path to OMBD folder is correct. """
    pass


class MonetaDir(object):
    def __init__(self):
        self.path = None

    def set_path(self, path: str):
        path = Path(path)
        if not path.is_dir():
            raise WrongOMBDDirError()
        self.path = path / 'xray' / 'beamtimes' / '2018_ESRF_SC4813_coherence' / 'id10' / 'data'

    def __call__(self, *args, **kwargs):
        if not self.path:
            raise MonetaDirNotSetError()
        return str(self.path)

    def _set_default_dir(self):
        self.set_path(_DEDFAULT_MONETA_DIR)


MONETA_DIR = MonetaDir()


def measure_speed(title: Optional[str] = None):
    """
    Decorator function that prints time in sec spent on a target function
    :param title: title to print before time
    """
    title = title or 'Time counting'

    def decorator(func):
        def wrapper(*args, **kwargs):
            print('\n', '*' * 30, '\n')
            print(title)
            start = time.perf_counter()
            res = func(*args, **kwargs)
            end = time.perf_counter()
            print()
            print('Performed in %.4f sec' % float(end - start))
            return res

        return wrapper

    return decorator


def get_zaptime_files(zaptime_number: int, info: Optional[bool] = False) -> List[str]:
    """
    Looks for f'zaptime_{zaptime_number}' folder in MONETA groupshare
    and returns the list of all /*.edf.gz file paths inside the folder, sorted by file number.

    :param zaptime_number: number of target folder
    :param info: if True, prints number of files found
    :return: list of file paths
    """
    pen_folders = get_listdir(MONETA_DIR(), start_filter='PEN')
    zaptime_name = '_'.join(['zaptime', str(zaptime_number)])
    for p_folder in pen_folders:
        if zaptime_name in os.listdir(p_folder):
            files = get_listdir(os.path.join(p_folder, zaptime_name), end_filer='.edf.gz')
            files.sort(key=lambda x: int(x.split('_')[-1].split('.')[0]))
            if info:
                logging.info('Found %d files in %s folder' % (len(files), zaptime_name))
            return files
    else:
        print('Folder %s is not found in %s' % (zaptime_name, MONETA_DIR()))
        return []


def get_pen_folder_number(zaptime_number: int) -> Optional[int]:
    if zaptime_number > 86:
        return 5
    pen_folders = get_listdir(MONETA_DIR(), start_filter='PEN')
    pen_folders.sort(key=lambda x: int(x.split('PEN')[-1]))
    zaptime_name = '_'.join(['zaptime', str(zaptime_number)])
    for i, p_folder in enumerate(pen_folders):
        if zaptime_name in os.listdir(p_folder):
            return i + 1
    else:
        print('Folder %s is not found in %s' % (zaptime_name, MONETA_DIR()))
        return


def get_zaptime_file(zaptime_number: int, file_num: int) -> Optional[str]:
    """
    Looks for f'zaptime_{zaptime_number}' folder in MONETA groupshare
    and returns the path of /*.edf.gz file with the corresponding number.

    :param zaptime_number: number of target folder
    :param file_num: number of target file
    :return: list of file paths
    """
    files = get_zaptime_files(zaptime_number)
    try:
        return files[file_num]
    except IndexError:
        logging.error('File number %d not found' % file_num)
        return


def get_listdir(dir_, *, start_filter: str = None, end_filer: str = None):
    """
    Lists files and folders in a target directory with optional name filters.
    :param dir_: target directory
    :type dir_: str
    :param start_filter:
    :param end_filer:
    :return:
    """
    start_filter = start_filter or ''
    end_filer = end_filer or ''
    if os.path.isdir(dir_):
        return [os.path.join(dir_, file_) for file_ in os.listdir(dir_)
                if file_.startswith(start_filter) and
                file_.endswith(end_filer)]
    else:
        raise NotADirectoryError('No such a directory: %s' % dir_)


def save_create(parent, group):
    """
    Creates group if doesn't exist and returns it

    :param parent: h5py.Group
    :param group: str
    :return: h5py.Group
    """
    if group not in list(parent.keys()):
        return parent.create_group(group)
    else:
        return parent[group]


def itersubclasses(cls, _seen=None):
    """
    itersubclasses(cls)

    Generator over all subclasses of a given class, in depth first order.

    """

    if not isinstance(cls, type):
        raise TypeError('itersubclasses must be called with '
                        'new-style classes, not %.100r' % cls)
    if _seen is None: _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError:  # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub


def get_image_header(folder_num: int, file_num: int) -> dict:
    """
    Returns edf header as a dict.
    :param folder_num: folder number
    :param file_num: file number
    :return: header dict
    """
    filepath = get_zaptime_file(folder_num, file_num)
    if filepath is None:
        raise FileNotFoundError('No file %d or folder %d found' % (file_num, folder_num))
    return read_header_from_file(filepath)


def setup_plot_fontsize(fontsize: int = 20,
                        bold: bool = True,
                        usetex: bool = True):
    import matplotlib as mpl
    from matplotlib import rc
    if bold:
        mpl.rcParams['text.latex.preamble'] = [r'\boldmath']
    if usetex:
        rc('text', usetex=True)

    rc('xtick', labelsize=fontsize)
    rc('ytick', labelsize=fontsize)
    rc('axes', labelsize=fontsize)
    rc('axes', titlesize=fontsize)
    rc('axes', titlesize=fontsize)
    rc('legend', fontsize=fontsize)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Contains MetaData class which allows to extract image properties such as temperatures and evaporation rates
"""

import os.path
from typing import List, Optional, Any
import logging

import h5py
import numpy as np

from gisaxs_xpcs.common_tools import get_zaptime_files


class MetaData(object):
    class Property(object):
        def __init__(self, h5name: str, unit: str = '', description: str = '',
                     name: str = None, value: Any = None, abs_time: float = None):
            self.h5name = h5name
            self.name = name or self.h5name
            self.unit = unit
            self.description = description
            self.value = value
            self.abs_time = abs_time

        def get_property(self, value, abs_time: float = None):
            return MetaData.Property(self.h5name, self.unit, self.description, self.name,
                                     value, abs_time)

        def __repr__(self):
            if isinstance(self.value, float):
                return '%s = %.2f (%s): %s. code: %s' % (self.name,
                                                         self.value, self.unit, self.description, self.h5name)
            else:
                return '%s (%s): %s. code: %s' % (self.name, self.unit, self.description, self.h5name)

    ###################################

    PROPERTIES = (Property('SThick', 'nm',
                           '''Film thickness estimated by crystal frequency changing''', 'thickness'),
                  Property('Temp1', '\N{DEGREE SIGN}C', 'PEN temperature of evaporation',
                           'temp_evaporation'),
                  Property('Temp2', '\N{DEGREE SIGN}C', 'Sample temperature', 'temp_sample'),
                  Property('ScalcRate60', 'nm/min', 'Smoothed rate of evaporation (1min average)',
                           'evaporation_rate'))

    ###################################

    class ImageProperties(dict):

        def __init__(self, *args, abs_time: float = None):
            if len(args) != len(MetaData.PROPERTIES):
                raise ValueError('Wrong number of properties passed: %d' % len(args))

            prop_dict = {property_.name: property_.get_property(value, abs_time)
                         for property_, value in zip(MetaData.PROPERTIES, args)}
            if abs_time:
                prop_dict['abs_time'] = abs_time

            super(MetaData.ImageProperties, self).__init__(**prop_dict)

        def __repr__(self):
            return '\n'.join([self.get(property_.name).__repr__() for property_ in MetaData.PROPERTIES])

    ###################################

    def __init__(self, filepath: str = None):
        self.filepath = filepath or os.path.join('metadata', 'metadata.h5')

    def get_image_properties(self, folder_num: int, file_num: int,
                             return_dict: bool = False) -> dict:
        try:
            with h5py.File(self.filepath, 'r') as f:
                common_group = f['common']
                zaptime_dset_name = 'zaptime_%d_abs_time' % folder_num
                image_abs_time = common_group[zaptime_dset_name][file_num]
                abs_time = common_group['abs_time'][()]
                index = np.argmin(np.abs(abs_time - image_abs_time))
                logging.debug('index = %d' % index)
                logging.debug('image_abs_time = %.2f' % image_abs_time)
                properties = [common_group[pr.h5name][index] for pr in self.PROPERTIES]
                if not return_dict:
                    return MetaData.ImageProperties(*properties, abs_time=image_abs_time)
                else:
                    return {pr.name: value for pr, value in zip(self.PROPERTIES, properties)}
        except Exception as err:
            print(err)
            return dict({'error': err})

    def get_first_and_last(self, folder_num: int) -> Optional[List[dict]]:
        files = get_zaptime_files(folder_num)
        if not files:
            return
        else:
            return [self.get_image_properties(folder_num, 0),
                    self.get_image_properties(folder_num, len(files) - 1)]

    def get_latex_title(self, folder_num: int, file_num: int,
                        attrs: dict = None) -> str:
        attrs = attrs or self.get_image_properties(folder_num, file_num)
        return '\n$T_{sample} = %.2f$ %s, $T_{pen} = %.2f$ %s, d = %.1f %s\n' % \
               (attrs['temp_sample'].value, attrs['temp_sample'].unit,
                attrs['temp_evaporation'].value, attrs['temp_evaporation'].unit,
                attrs['thickness'].value, attrs['thickness'].unit)


if __name__ == '__main__':
    from pprint import pprint

    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s:%(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p')

    metadata = MetaData()
    # pprint([p['temp_sample'] for p in metadata.get_first_and_last(16)])
    pprint(metadata.get_image_properties(19, 1300))



# -*- coding: utf-8 -*-

from gisaxs_xpcs import *


def show_folders_info(folder_list: List[int]):
    print('\n'.join([': '.join([str(f), str(len(get_zaptime_files(f))), f'PEN0{get_pen_folder_number(f)}'])
                     for f in folder_list]))


def show_folder_first_and_last(folder_num):
    first, last = ed0.metadata.get_first_and_last(folder_num)
    print('First frame: ', first, sep='\n\n')
    print('-' * 50)
    print('Last frame: ', last, sep='\n\n')


def get_all_files(folder_num_list, max_number_of_files_in_each_folder=None):
    files = list()
    for folder_num in folder_num_list:
        number_of_files = len(get_zaptime_files(folder_num)) - 2
        step = 1 if not max_number_of_files_in_each_folder else int(
            number_of_files / max_number_of_files_in_each_folder)
        files.extend([(folder_num, i) for i in range(1, number_of_files, step)])
    return files


def get_cmap(y):
    y = np.array(y)
    norm = mpl.colors.Normalize(vmin=y.min(), vmax=y.max())
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.jet)
    cmap.set_array([])
    return cmap


def get_nm(delta):
    """
    Returns size in [nm] corresponding for the provided delta in degrees
    for 1.5307 [A] wavelength (common approximate formula is lambda [nm] / delta [rad] for small angles).
    """
    return 8.77 / delta


def get_intensity_from_moneta(files, **kwargs):
    intensity_array = np.empty(len(files))
    for i, f in enumerate(tqdm(files)):
        df = ed0.get_images(*f, **kwargs)
        intensity_array[i] = df.sum().sum()
    time_axis = get_time_axis(files)
    plt.plot(time_axis, intensity_array)
    ax = plt.gca()
    ax.set_xlabel('Time (sec)')
    ax.set_ylabel('Integrated intensity of the frame (a.u.)')


def get_intensity_from_h5(files):
    h5_filepath = os.path.join('metadata', 'test_raw_data.h5')

    def dset_path(file: Tuple[int, int], z_number=4):
        return '/'.join(['Raw data', 'zaptime_%d' % file[0], str(file[1]).zfill(z_number)])

    def group_path(file: Tuple[int, int]):
        return '/'.join(['Raw data', 'zaptime_%d' % file[0]])

    intensity_array = np.empty(len(files))

    with h5py.File(h5_filepath) as h5file:
        for i, f in enumerate(tqdm(files)):
            z_number = len(list(h5file[group_path(f)].keys())[0])
            intensity_array[i] = h5file[dset_path(f, z_number)][()].flatten().sum()

    time_axis = get_time_axis(files)
    plt.plot(time_axis, intensity_array)
    ax = plt.gca()
    ax.set_xlabel('Time (sec)')
    ax.set_ylabel('Integrated intensity of the frame (a.u.)')


def plot_metadata(files: List[Tuple[int, int]], metadata_key):
    prop = ed1.metadata.get_image_properties(*files[0], return_dict=False)

    properties = [ed1.metadata.get_image_properties(*f, return_dict=True)[metadata_key] for f in files]
    time_axis = get_time_axis(files)
    plt.plot(time_axis, properties, '.-')
    ax = plt.gca()
    ax.set_xlabel('Time (sec)')
    ylabel = prop[metadata_key].description if metadata_key != 'thickness' else 'Thickness'
    ax.set_ylabel(fr'{ylabel}, {prop[metadata_key].unit}')
    ax.grid(linestyle='--', linewidth=0.5)


def get_files_list(folder_num, num):
    folder_size = len(get_zaptime_files(folder_num))
    return [(folder_num, int(i)) for i in np.linspace(0, folder_size - 2, num)]


# def get_time_axis(files: List[Tuple[int, int]]):
#     def func(n):
#         try:
#             return ed0.metadata.get_image_properties(*n)['abs_time']
#         except KeyError as err:
#             print('Folder, file : ', n)
#             raise KeyError(err)
#     start = func(files[0])
#     return [func(n) - start for n in files]

# def get_temp_axis(files: List[Tuple[int, int]], key='temp_sample'):
#     def func(n):
#         try:
#             return ed0.metadata.get_image_properties(*n)[key].value
#         except KeyError as err:
#             print('Folder, file : ', n)
#             raise KeyError(err)
#     return [func(n) for n in files]

def show_roi(folder_number, file_number,
             x_region, y_region, gp, **kwargs):
    cut_x = Cut(x_region, axis=0)
    cut_y = Cut(y_region, axis=1)
    plot = PlotReport(folder_number, file_number, cut_x=cut_x,
                      cut_y=cut_y, gp=gp, **kwargs)

    fig3 = plt.figure(figsize=(30, 25))
    gs = fig3.add_gridspec(3, 3)
    f3_ax1 = fig3.add_subplot(gs[0:2, 1])
    f3_ax2 = fig3.add_subplot(gs[0, 0])
    f3_ax3 = fig3.add_subplot(gs[1, 0])

    plt.sca(f3_ax2)
    cut_x.plot_cut(df=plot.df)
    plt.sca(f3_ax3)
    cut_y.plot_cut(df=plot.df)
    plt.sca(f3_ax1)
    plot.plot_im(ax=f3_ax1)
    plt.subplots_adjust(hspace=0.25)


def get_cut_from_df(df, cut, axis, sigma=0, **kwargs):
    df = df.iloc[(df.index > cut[0]) * (df.index < cut[1]),
                 (df.columns > cut[2]) * (df.columns < cut[3])]

    axes = (df.columns, df.index)
    y = df.mean(axis=axis)
    if sigma > 0:
        y = gaussian_filter1d(y, sigma=sigma)
    x = axes[axis]
    return x, y


def get_cut(folder_num, file_num, cut, axis, sigma=0, **kwargs):
    df = ExtractData().get_images(folder_num, file_num,
                                  plot=False, **kwargs)

    return get_cut_from_df(df, cut, axis, sigma=sigma, **kwargs)


def compare_cuts(files: List[Tuple[int, int]], cut, axis, sigma=0, *,
                 substrate_file: Tuple[int, int] = None, logscale=True,
                 temp_axis_key=None, figsize=(7, 7), title: str = None, **kwargs):
    cut_list = list()

    if substrate_file is not None:
        x, substrate = get_cut(*substrate_file, cut, axis, sigma, **kwargs)

    if temp_axis_key is None:
        time_axis = get_time_axis(files)
    else:
        time_axis = get_temp_axis(files, temp_axis_key)
    cmap = get_cmap(time_axis)

    plt.figure(figsize=figsize)
    ax = plt.gca()
    for t, f in zip(time_axis, files):
        x, y = get_cut(*f, cut, axis, sigma, **kwargs)
        if substrate_file is not None:
            y -= substrate
        cut_list.append(y)
        if logscale:
            ax.semilogy(x, y, color=cmap.to_rgba(t))
        else:
            ax.plot(x, y, color=cmap.to_rgba(t))

    if axis == 0:
        ax.set_xlabel(r'$2 \theta$, deg')
    else:
        ax.set_xlabel(r'$\alpha$, deg')
    if logscale:
        ax.set_ylabel(r'$log(I)$')
    else:
        ax.set_ylabel(r'$I$')
    ax.grid(linestyle='--', linewidth=0.5)
    cbar = plt.colorbar(cmap)
    if temp_axis_key is None:
        cbar_label = 'Time (sec)'
    else:
        prop = ExtractData().metadata.get_image_properties(*files[0])[temp_axis_key]
        if temp_axis_key == 'thickness':
            prop.description = 'Thickness'
        cbar_label = f'{prop.description}, {prop.unit}'

    cbar.set_label(cbar_label)

    if title is not None:
        plt.title(title)
    elif axis == 0:
        horizontal_title(cut)
    else:
        vertical_title(cut)

    return cut_list


def plot_gisaxs(df, ax=None, colorbar=True):
    ax = ax if ax is not None else plt.gca()
    extent = ed1.get_extent(df)
    im = ax.imshow(df, extent=extent, cmap='jet')
    ax.set_xlabel(r'$2 \theta$, deg')
    ax.set_ylabel(r'$\alpha$, deg')
    if colorbar:
        cbar = plt.colorbar(im, fraction=0.046, pad=0.04)


#         cbar.set_label(r'$log(I)$')

def log_df(df, background=0.01):
    return np.log(np.clip(df, background, np.amax(df.values)))


def gisaxs_difference(file: Tuple[int, int], substrate_file: Tuple[int, int], plot=True, **kwargs):
    df = ed1.get_images(*file, apply_log=False, **kwargs)
    df_substrate = ed1.get_images(*substrate_file, apply_log=False, **kwargs)
    df_clean = df - df_substrate
    df_clean = log_df(df_clean)
    if plot:
        plt.figure(figsize=(10, 10))
        plot_gisaxs(df_clean)
    return df_clean


def get_rectangle(cut, props):
    return patches.Rectangle((cut[2], cut[0]), cut[3] - cut[2], cut[1] - cut[0], **props)


def get_title(file, title=None):
    title = title or 'Pentacene sample %d' % get_pen_folder_number(file[0])
    try:
        title = '%s%s(folder %d, image %d)' % (
            title,
            ed1.metadata.get_latex_title(*file),
            *file)
    except Exception as err:
        pass
    return title


def show_roi_diff(file: Tuple[int, int], substrate_file: Tuple[int, int],
                  x_region, y_region, **kwargs):
    sigma = kwargs.pop('sigma', 2)
    title = kwargs.pop('title', None)
    title = get_title(file, title)

    cut_x = Cut(x_region, axis=0, sigma=sigma)
    cut_y = Cut(y_region, axis=1, sigma=sigma)
    rectangle_props = dict(fill=False, color='red', alpha=1, lw=3, ls='--')

    rectangle_props.update(kwargs.pop('rect', dict()))

    df_original = ed1.get_images(*file, apply_log=True, background=0.1, **kwargs)
    df = gisaxs_difference(file, substrate_file, plot=False, **kwargs)

    fig3 = plt.figure(figsize=(30, 25))
    gs = fig3.add_gridspec(3, 3)
    f3_ax1 = fig3.add_subplot(gs[0:2, 1])
    f3_ax2 = fig3.add_subplot(gs[0, 0])
    f3_ax3 = fig3.add_subplot(gs[1, 0])

    r1 = get_rectangle(x_region, rectangle_props)
    r2 = get_rectangle(y_region, rectangle_props)

    plt.sca(f3_ax2)
    cut_x.plot_cut(df=df)
    plt.sca(f3_ax3)
    cut_y.plot_cut(df=df)
    plt.sca(f3_ax1)
    plot_gisaxs(df_original)
    f3_ax1.add_patch(r1)
    f3_ax1.add_patch(r2)
    plt.subplots_adjust(hspace=0.25)
    plt.title(title)


def get_simple_title(file):
    attrs = ed1.metadata.get_image_properties(*file)
    return '$T_{sample} = %.2f$ %s, $T_{pen} = %.2f$ %s, $d = %.1f$ %s\n(folder %d, image %d)' % \
           (attrs['temp_sample'].value, attrs['temp_sample'].unit,
            attrs['temp_evaporation'].value, attrs['temp_evaporation'].unit,
            attrs['thickness'].value, attrs['thickness'].unit,
            *file)


def get_very_simple_title(file, key):
    latex_dict = dict(temp_sample='$T_{sample} = %.2f$ %s',
                      temp_evaporation='$T_{pen} = %.2f$ %s',
                      thickness='$d = %.1f$ %s')
    attrs = ed1.metadata.get_image_properties(*file)

    title = latex_dict[key] % (attrs[key].value, attrs[key].unit)
    return title


def four_gisaxs_pics(files: List[Tuple[int, int]], figsize=(9, 9), key=None, regions=None, **kwargs):
    assert len(files) == 4, 'It is 4 GISAXS pics, man!'

    default_params = dict(apply_log=True, background=0.1,
                          x_range=(-0.2, 0.2), y_range=(0, 0.4))

    rectangle_props = dict(fill=False, color='red', alpha=1, lw=3, ls='--')
    rectangle_props.update(kwargs.pop('rect', dict()))
    default_params.update(kwargs)

    fig, axs = plt.subplots(2, 2, sharey=False, figsize=figsize)
    axs = axs.flatten()
    for i, (ax, file) in enumerate(zip(axs, files)):
        df = ed1.get_images(*file, plot=False, **default_params)
        if i == 1 or i == 3:
            ax.yaxis.tick_right()
        if i == 0 or i == 1:
            ax.xaxis.tick_top()
        extent = ed1.get_extent(df)
        im = ax.imshow(df, extent=extent, cmap='jet')
        if i == 2 or i == 3:
            ax.set_xlabel(r'$2 \theta$, deg')
        if i == 0 or i == 2:
            ax.set_ylabel(r'$\alpha$, deg')

        if key:
            label = get_very_simple_title(file, key)
            text = ax.text(0.5, 0.85, label,  # r'$T_$ nm', #% get_very_simple_title(file, key),
                           horizontalalignment='center',
                           verticalalignment='center',
                           transform=ax.transAxes, color='white', fontsize=18, weight='extra bold')
            text.set_bbox(dict(facecolor='black', alpha=0.65, edgecolor='black'))

        if regions is not None:
            for cut in regions:
                ax.add_patch(get_rectangle(cut, rectangle_props))

    plt.subplots_adjust(hspace=0, wspace=0.2)


def one_gisaxs_pic(file: Tuple[int, int], figsize=(9, 9), regions=None,
                   title: str = False, axis_off: bool = True, **kwargs):
    default_params = dict(apply_log=True, background=0.1,
                          x_range=(-0.2, 0.2), y_range=(0, 0.4))

    rectangle_props = dict(fill=False, color='red', alpha=1, lw=3, ls='--')
    rectangle_props.update(kwargs.pop('rect', dict()))
    rectangles = list()

    if regions is not None:
        for cut in regions:
            rectangles.append(get_rectangle(cut, rectangle_props))
    default_params.update(kwargs)

    fig = plt.figure(figsize=figsize)
    ax = plt.gca()
    df = ed1.get_images(*file, plot=False, **default_params)
    plot_gisaxs(df, ax, False)
    if title:
        ax.set_title(get_simple_title(file))
    for r in rectangles:
        ax.add_patch(r)
    plt.tight_layout()
    if axis_off:
        ax.set_axis_off()


def vertical_title(cut):
    plt.title(fr'Vertical cuts ' \
                  fr'$2 \theta = {cut[2]} .. {cut[3]}$ deg');


def horizontal_title(cut):
    plt.title(fr'Horizontal cuts ' \
                  fr'$\alpha = {cut[0]} .. {cut[1]}$ degree');


def plot_cut_fit(cut, p0=None, fit=True):
    """
    If fit is True, fits the cut with initial p0
    (or with predefined initial point if p0 is None) and plots the result of fitting.

    If fit is False, just plots cut with initial fitting function.
    """
    if fit:
        cut.fit.fit(x=cut.x, y=cut.y, p0=p0)
        p0 = cut.fit.fit_res[0]
    plt.figure(figsize=(10, 5))
    cut.fit.plot(x=cut.x, y=cut.y, args=p0)
    ax = plt.gca()
    ax.set_ylim([np.min(cut.y) * 0.95, np.max(cut.y) * 1.05])


def plot_simple_cut(file: Tuple[int, int],
                    cut: Tuple[float, float, float, float],
                    axis: int,
                    *,
                    sigma: float = 0,
                    logscale: bool = True,
                    title: str = None,
                    figsize: Tuple[int, int] = (8, 8)
                    ):
    df = ExtractData().get_images(*file, apply_log=False, x_range=cut[:2],
                                  y_range=cut[2:], plot=False)
    y = df.mean(axis=axis)
    x = (df.columns, df.index)[axis]
    if sigma > 0:
        y = gaussian_filter1d(y, sigma)
    plt.figure(figsize=figsize)
    if logscale:
        plt.semilogy(x, y)
    else:
        plt.plot(x, y)

    ax = plt.gca()
    if axis == 0:
        ax.set_xlabel(r'$2 \theta$, deg')
    else:
        ax.set_xlabel(r'$\alpha$, deg')
    if logscale:
        ax.set_ylabel(r'$log(I)$')
    else:
        ax.set_ylabel(r'$I$')
    ax.grid(linestyle='--', linewidth=0.5)

    if title is not None:
        plt.title(title)
    elif axis == 0:
        horizontal_title(cut)
    else:
        vertical_title(cut)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Contains ExtractData class which allows to extract and process data
from the remote server and .h5 files
(images data, plots, movies, metadata)
"""

from typing import List, Tuple, Union, Optional
import warnings

import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib import rc

from gisaxs_xpcs.common_tools import get_zaptime_files, get_pen_folder_number
from gisaxs_xpcs.read_edf import read_edf_gz
from gisaxs_xpcs.gisaxs_parameters import GisaxsParameters
from gisaxs_xpcs.metadata_obj import MetaData

rc('text', usetex=True)
rc('xtick', labelsize=12)
rc('ytick', labelsize=12)
rc('axes', labelsize=14)


class ExtractData(object):
    def __init__(self, *,
                 default_background: float = 0.01,
                 metadata: MetaData = None):

        self.default_background = default_background
        self.metadata = metadata or MetaData()

    @staticmethod
    def zaptime_info(folder_num: int):
        return 'Number of files = %d' % len(get_zaptime_files(folder_num))

    # @measure_speed('Getting images ... ')
    def get_images(self, folder_num: int, num: Optional[Union[int, list]] = None, *,
                   plot: bool = False,
                   apply_log: bool = False,
                   background: float = None,
                   x_range: tuple = (-0.1, 0.1),
                   y_range: tuple = (0., 0.4),
                   gp: GisaxsParameters = None,
                   vmin: float = None,
                   vmax: float = None,
                   get_properties: bool = False,
                   save: bool = False,
                   title: Optional[str] = None,
                   show: bool = True) -> DataFrame or List[DataFrame]:

        background = background or self.default_background
        gp = gp or GisaxsParameters.init_from_folder_number(folder_num)
        files = get_zaptime_files(folder_num)
        if num is None:
            num = list(range(len(files)))
        if isinstance(num, list):
            x_angles, y_angles = gp.get_angle_vectors(units='deg', flip=False)
            res = []
            for n in num:
                df = self.get_df_from_data(read_edf_gz(files[n])[0], self.get_name(folder_num, n),
                                           apply_log,
                                           background, x_range, y_range, gp,
                                           x_angles, y_angles)
                if get_properties:
                    df.attrs.update(self.metadata.get_image_properties(folder_num, n, return_dict=True))
                if plot or save:
                    self._plot_image(df, vmin, vmax, show=show, save=save, title=title, folder_num=folder_num, num=n)
                res.append(df)
            return res
        else:
            file_ = files[num] if files and len(files) > num + 1 else None
            if not file_:
                raise ValueError('num = %d is bigger than number of files = %d' % (num, len(files)))
            data, header_dict = read_edf_gz(files[num])
            df = self.get_df_from_data(data, self.get_name(folder_num, num),
                                       apply_log, background, x_range, y_range, gp)
            if plot or save:
                self._plot_image(df, vmin, vmax, show=show, save=save, title=title,
                                 folder_num=folder_num, num=num)
            if get_properties:
                df.attrs.update(self.metadata.get_image_properties(folder_num, num))
            return df

    @staticmethod
    def get_name(folder_num: int, num: int) -> str:
        return 'image_%d_%d' % (folder_num, num)

    @staticmethod
    def get_df_from_data(data, name: str, apply_log: bool, background: float,
                         x_range: tuple, y_range: tuple, gp: GisaxsParameters,
                         x_angles: np.array = None, y_angles: np.array = None) -> DataFrame:
        if x_angles is None or y_angles is None:
            x_angles, y_angles = gp.get_angle_vectors(units='deg', flip=False)
            x_angles = np.flip(x_angles, axis=0)
        df = DataFrame(np.flip(np.clip(data, background, np.amax(data)), axis=1),
                       columns=x_angles, index=y_angles)
        df = df.iloc[(df.index >= y_range[0]) * (df.index <= y_range[1]),
                     (df.columns >= x_range[0]) * (df.columns <= x_range[1])]
        if apply_log:
            df = np.log(df)
        df.name = name
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df.attrs = dict()
        return df

    # def _str_from_attrs(self, attrs: dict) -> str:
    #     def _chech_class(value) -> float:
    #         if not isinstance(value, float):
    #             return value.value
    #         else:
    #             return value
    #     props = 0

    def _plot_image(self, df: DataFrame, vmin: float = None, vmax: float = None,
                    save: bool = False, path: str = None, title: str = None,
                    folder_num: int = None, num: int = None, extent_: tuple = None,
                    show: bool = True):

        fig, ax = plt.subplots()
        extent_ = extent_ or self.get_extent(df)
        p = ax.imshow(df, extent=extent_,
                      vmin=vmin,
                      vmax=vmax,
                      cmap='jet')
        if folder_num and num:
            if title is None:
                title = 'Pentacene sample %d' % get_pen_folder_number(folder_num)
            try:
                title = '%s%s(folder %d, image %d)' % (
                    title,
                    self.metadata.get_latex_title(folder_num, num),
                    folder_num, num)
            except Exception as err:
                print(err)
            plt.title(title)
        ax.set_xlabel(r'$2 \theta$, deg')
        ax.set_ylabel(r'$\alpha$, deg')
        plt.colorbar(p, ax=ax)
        if save:
            path = path or '%s.png' % df.name
            plt.savefig(path)
        elif show:
            plt.show()
        return p

    @staticmethod
    def get_extent(df, numbers: int = 2):
        return (round(df.columns.min(), numbers),
                round(df.columns.max(), numbers),
                round(df.index.min(), numbers),
                round(df.index.max(), numbers),)

    def make_video(self, folder_num: int, *, background: float = 0.1,
                   x_range: tuple = (-0.1, 0.1), y_range: tuple = (0., 0.4),
                   vmin: float = None, vmax: float = None, apply_log=True,
                   number_of_frames: int = 100,
                   interval: int = 200, title: str = None) -> None:

        files = get_zaptime_files(folder_num)

        def df_list(num: int) -> DataFrame:
            return self.get_images(folder_num, num, apply_log=apply_log, background=background,
                                   x_range=x_range, y_range=y_range)

        print('%d frames found ' % len(files))
        if len(files) == 0:
            return

        number_of_frames = number_of_frames or len(files)
        number_of_frames = min([number_of_frames, len(files)])

        delta_frame = int(len(files) / number_of_frames)

        extent_ = self.get_extent(df_list(0))

        fig = plt.figure()
        if title is None:
            title = 'Pentacene sample %d' % get_pen_folder_number(folder_num)

        def update_imshow(num):
            num = num * delta_frame if num * delta_frame < len(files) else len(files) - 1
            print(num)
            p = plt.imshow(df_list(num), vmin=vmin, vmax=vmax, extent=extent_)
            ax = plt.gca()
            ax.set_xlabel(r'$2 \theta$, deg')
            ax.set_ylabel(r'$\alpha$, deg')
            if not update_imshow.colorbar_is_set:
                plt.colorbar()
                update_imshow.colorbar_is_set = True
            plt.title('%s%s(folder %d, image %d)' % (
                title,
                self.metadata.get_latex_title(folder_num, num),
                folder_num, num))
            return p,

        update_imshow.colorbar_is_set = False
        ani = animation.FuncAnimation(fig, update_imshow, number_of_frames,
                                      interval=interval, blit=True)

        ani.save('movie_%d.mp4' % folder_num)


def get_time_axis(files: List[Tuple[int, int]]):
    def func(n):
        try:
            return MetaData().get_image_properties(*n)['abs_time']
        except KeyError as err:
            print('Folder, file : ', n)
            raise KeyError(err)

    start = func(files[0])
    return [func(n) - start for n in files]


def get_temp_axis(files: List[Tuple[int, int]], key='temp_sample'):
    def func(n):
        try:
            return MetaData().get_image_properties(*n)[key].value
        except KeyError as err:
            print('Folder, file : ', n)
            raise KeyError(err)

    return [func(n) for n in files]


if __name__ == '__main__':
    from pprint import pprint

    extract_data = ExtractData()
    folder_number = 99
    print(extract_data.zaptime_info(folder_number))
    ims = extract_data.get_images(folder_number, [1], background=1,
                                  apply_log=True, plot=True, get_properties=True,
                                  save=False)
    for im in ims:
        pprint(im.attrs)

    # extract_data.make_video(97, number_of_frames=100, title='Pentacene sample 5, annealing')



# !/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import gzip

import numpy as np

__all__ = ['read_edf', 'read_edf_header',
           'read_edf_from_data', 'read_edf_gz', 'read_header_from_file', 'read_edf_from_file']


def read_edf_from_file(file_path: str):
    data = get_data_from_filepath(file_path)
    return read_edf_from_data(data)


def read_edf_gz(gz_filepath, *, reshape: bool = True):
    _check_file(gz_filepath, '.edf.gz')
    with gzip.open(gz_filepath, 'rb') as f:
        data = f.read()
    return read_edf_from_data(data, reshape=reshape)


def read_edf(edf_filepath, *, reshape: bool = True):
    _check_file(edf_filepath, '.edf')
    with open(edf_filepath, 'rb') as f:
        data = f.read()
    return read_edf_from_data(data, reshape=reshape)


def read_edf_from_data(data, *, reshape: bool = True):
    header_dict = read_header_from_data(data)
    header_end_index = header_dict['headerSize']
    image_size = int(header_dict['Size'])
    raw_image_data = data[header_end_index:header_end_index + image_size]
    data_type = _get_numpy_type(header_dict['DataType'])
    image_shape = (int(header_dict['Dim_2']), int(header_dict['Dim_1']))

    data = np.frombuffer(raw_image_data, data_type)
    if reshape:
        data = np.rot90(np.reshape(data, image_shape))
    return data, header_dict


def read_edf_header(edf_filepath):
    _check_file(edf_filepath, '.edf')
    with open(edf_filepath, 'rb') as f:
        data = f.read()
    return read_header_from_data(data)


def read_edf_header_from_gz(gz_filepath):
    _check_file(gz_filepath, '.edf.gz')
    with gzip.open(gz_filepath, 'rb') as f:
        data = f.read()
    return read_header_from_data(data)


def read_header_from_data(data) -> dict:
    header_end_index = data.find(b'}\n') + 2
    if header_end_index != 1024:
        print('File has unusual size of header bites %i' % header_end_index)
    header = data[1:header_end_index].decode('utf-8')
    header_dict = _get_header_dict(header)
    header_dict.update({'headerSize': header_end_index})
    return header_dict


def read_header_from_file(filepath):
    data = get_data_from_filepath(filepath)
    return read_header_from_data(data)


def get_data_from_filepath(filepath: str):
    _check_file(filepath)
    if filepath.endswith('.edf'):
        with open(filepath, 'rb') as f:
            return f.read()
    elif filepath.endswith('.edf.gz'):
        with gzip.open(filepath, 'rb') as f:
            return f.read()
    else:
        raise ValueError('Unknown file type')


def _get_header_dict(header):
    header_dict = {}
    raw_list = header.replace('\n', '').strip(). \
        replace(' ', ''). \
        replace('{', ''). \
        replace('}', ''). \
        split(';')
    for item in raw_list:
        item = item.split('=')
        if len(item) == 2:
            header_dict.update([item])
    return header_dict


def _check_file(filepath: str, end_filter: str = None) -> None:
    end_filter = end_filter or ''
    if not os.path.isfile(filepath):
        raise FileNotFoundError('File %s doesn\'t exist' % filepath)
    if not filepath.endswith(end_filter):
        raise ValueError('File %s is not an edf.gz file' % filepath)


def _get_numpy_type(edf_type):
    """
    Returns NumPy type based on edf type
    """
    edf_type = edf_type.upper()
    if edf_type == "SIGNEDBYTE":
        return np.int8  # "b"
    elif edf_type == "UNSIGNEDBYTE":
        return np.uint8  # "B"
    elif edf_type == "SIGNEDSHORT":
        return np.int16  # "h"
    elif edf_type == "UNSIGNEDSHORT":
        return np.uint16  # "H"
    elif edf_type == "SIGNEDINTEGER":
        return np.int32  # "i"
    elif edf_type == "UNSIGNEDINTEGER":
        return np.uint32  # "I"
    elif edf_type == "SIGNEDLONG":
        return np.int32  # "i"
    elif edf_type == "UNSIGNEDLONG":
        return np.uint32  # "I"
    elif edf_type == "SIGNED64":
        return np.int64  # "l"
    elif edf_type == "UNSIGNED64":
        return np.uint64  # "L"
    elif edf_type == "FLOATVALUE":
        return np.float32  # "f"
    elif edf_type == "FLOAT":
        return np.float32  # "f"
    elif edf_type == "DOUBLEVALUE":
        return np.float64  # "d"
    else:
        raise TypeError("unknown EdfType %s" % edf_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import re
import logging
from typing import Tuple, Union, List
from pprint import pprint

import numpy as np
from scipy.ndimage import zoom
import h5py
from tqdm import tqdm

import matplotlib.pyplot as plt
import matplotlib.patches as patches

from gisaxs_xpcs.common_tools import get_zaptime_files
from gisaxs_xpcs.gisaxs_parameters import GisaxsParameters
from gisaxs_xpcs.extract_data import ExtractData

logger = logging.getLogger(__name__)


class XPCS(object):
    """
    Class contains the basic functionality to calculate
    two-time correlation functions for
    GISAXS images stored to .h5 file.
    """

    DEFAULT_H5_FILEPATH = '/media/vladimir/data/GISAXS_DATA/test_raw_data.h5'

    @staticmethod
    def _default_h5_path(folder_num: int):
        return '/'.join(['Raw data', 'zaptime_%s' % str(folder_num).zfill(2)])

    def __init__(self, folder_num: int = None,
                 cut_region: Tuple[float, float, float, float] = None,
                 gp: GisaxsParameters = None,
                 h5_filepath: str = None,
                 h5_path: str = None):
        """

        :param folder_num: number of the zaptime folder to analyse
        :param cut_region: tuple(x_min, x_max, y_min, y_max) in [deg]
        :param gp: gisaxs parameters object (necessary to extract angles)
        :param h5_filepath: path to the data source
        :param h5_path: path to zaptime folder in h5 file
        """
        if not folder_num and not h5_filepath:
            raise ValueError('Either folder_num or h5_filename should be provided')

        self.folder_num = folder_num
        self.size = len(get_zaptime_files(self.folder_num))
        self.index_x = self.index_y = None
        if not gp and folder_num:
            self.gp = GisaxsParameters.init_from_folder_number(folder_num)
        else:
            self.gp = gp or GisaxsParameters()
        self.cut_region = cut_region
        self.h5_filepath = h5_filepath or self.DEFAULT_H5_FILEPATH
        self.h5_path = h5_path or self._default_h5_path(self.folder_num)
        if not self.setup_h5():
            raise NotImplementedError('Reading from remote folder is not'
                                      ' implemented yet.')

        self.x_angles, self.y_angles = self.gp.get_angle_vectors(units='deg',
                                                                 flip=False)
        self.x, self.y = np.sort(self.x_angles), np.flip(np.sort(self.y_angles),
                                                         axis=0)

        if self.cut_region is not None:
            self.init_indexes()

    def set_region(self, cut_region: Tuple[float, float, float, float]):
        self.cut_region = cut_region
        self.init_indexes()

    def setup_h5(self) -> bool:
        """
        Checks .h5 file. If it exists and contains the provided path to the dataset,
        returns True, otherwise False.
        """
        if not os.path.isfile(self.h5_filepath):
            logger.error('File %s doesn\'t exist.' % self.h5_filepath)
            return False
        try:
            with h5py.File(self.h5_filepath, 'r') as f:
                if f.get(self.h5_path, None) is None:
                    logger.error('File does not contain path %s' % self.h5_path)
                    return False
                # data = f[self.h5_path][list(f[self.h5_path].keys())[0]][()]
                # print(data.shape)
        except Exception as err:
            logger.error(err)
            return False
        return True

    def init_indexes(self):
        self.index_x = np.min(np.where(self.y < self.cut_region[3])[0]), \
                       np.max(np.where(self.y > self.cut_region[2])[0])
        self.index_y = np.min(np.where(self.x > self.cut_region[0])[0]), \
                       np.max(np.where(self.x < self.cut_region[1])[0])
        logger.debug(f'self.index_x, self.index_y')

    def get_qt_array(self, mode='one-time'):
        if mode not in ['one-time', 'mean', 'std']:
            raise ValueError("Mode should be one of 'one-time', 'mean', 'std'.")
        if self.cut_region is None:
            raise ValueError('cut region is not defined')
        with h5py.File(self.h5_filepath, 'r') as f:
            digits_pattern = r'^([\s\d]+)$'
            dset_keys = [k for k in f[self.h5_path].keys() if re.search(digits_pattern, k)]
            dset_sorted_keys = sorted(dset_keys, key=lambda x: int(x))

            dset = f[self.h5_path][dset_sorted_keys[0]]
            intensity = dset[self.index_x[0]:self.index_x[1],
                        self.index_y[0]:self.index_y[1]]

            intensity_array = np.empty((self.size, intensity.shape[0] * intensity.shape[1]))
            t_array = np.empty(self.size)

            for dset_number, dset_name in enumerate(dset_sorted_keys):
                dataset = f[self.h5_path][dset_name]
                intensity = dataset[self.index_x[0]:self.index_x[1],
                            self.index_y[0]:self.index_y[1]]
                intensity = intensity.flatten()

                if mode == 'one-time':
                    intensity_array[dset_number] = intensity
                elif mode == 'mean':
                    intensity_array[dset_number, :] = intensity / intensity.mean()
                else:
                    intensity_array[dset_number, :] = intensity / intensity.std()

                t_array[dset_number] = dataset.attrs.get('time_of_frame', dset_number)
            return intensity_array, t_array

    def get_one_time_corr_function(self, mode='one-time'):
        if mode not in ['one-time', 'mean']:
            raise ValueError('Mode should be one of "one-time" and "mean".')
        intensity_array, t_array = self.get_qt_array(mode=mode)
        if mode == 'one-time':
            norm_constant = intensity_array.mean(axis=1).mean(axis=0) ** 2
        else:
            norm_constant = 1
        corr_matrix = intensity_array.dot(intensity_array.transpose()) / intensity_array.shape[1]
        corr_function = np.empty(intensity_array.shape[0])

        for t in range(corr_matrix.shape[0]):
            corr_function[t] = corr_matrix.diagonal(t).mean() / norm_constant
        return corr_function, t_array

    def get_two_time_corr_function(self, mode='std'):
        if mode not in ['mean', 'std']:
            raise ValueError("Mode should be one of 'mean', 'std'.")
        intensity_array, _ = self.get_qt_array(mode=mode)
        if mode == 'mean':
            res = intensity_array.dot(intensity_array.transpose()) / intensity_array.shape[1]
        else:
            res = intensity_array.dot(intensity_array.transpose()) / intensity_array.shape[1]
            i_mean = np.expand_dims(intensity_array.mean(axis=1), axis=1)
            # assert i_mean.shape == (intensity_array.shape[0], 1), 'Wrong i_mean shape: %s' % str(i_mean.shape)
            # m_res = i_mean.dot(i_mean.transpose())
            # assert m_res.shape == res.shape, 'Wrong m_res shape: %s' % str(m_res.shape)
            res -= i_mean.dot(i_mean.transpose())
        return res

    def get_intensity(self):
        intensity, t = self.get_qt_array('one-time')
        intensity = intensity.mean(axis=1)
        return intensity, t

    def plot_cut(self, filenum: int = 1):
        extract_data = ExtractData()
        extract_data.get_images(self.folder_num, filenum,
                                apply_log=True, plot=True, show=False, gp=self.gp)
        ax = plt.gca()
        cut = self.cut_region
        r = patches.Rectangle((cut[0], cut[2]), cut[1] - cut[0], cut[3] - cut[2],
                              fill=False, color='red')
        ax.add_patch(r)
        plt.show()


class CutList(list):
    """
    CutList is a list containing elements
    of type Tuple[x_min, x_max, y_min, y_max].
    x_min, x_max, y_min, y_max are values in degrees
    that define a window on a GISAXS image within which
    XPCS correlation functions will be calculated.

    Should be only initialized via classmethods and never from __init__().
    """

    __slots__ = ['parameters']

    @classmethod
    def get_cut_list(cls, init_window: Union[Tuple[float, float, float, float],
                                             Tuple[float, float]],
                     axis: int, step: float, number_of_cuts: int):
        self = cls()
        if axis not in (0, 1):
            raise ValueError('Axis value should be 0 (cuts along q parallel)'
                             ' or 1 (cuts along q z axis)')
        if len(init_window) == 2:
            init_window = (init_window[0], init_window[0] + step,
                           init_window[1], init_window[1] + step)
        if axis == 0:
            self.append([(init_window[0] + step * i,
                          init_window[1] + step * i,
                          init_window[2],
                          init_window[3]) for i in range(number_of_cuts)])
        else:
            self.append([(init_window[0],
                          init_window[1],
                          init_window[2] + step * i,
                          init_window[3] + step * i)
                         for i in range(number_of_cuts)])
        setattr(self, 'parameters', dict(init_window=init_window, step=step,
                                         axis=axis, number_of_cuts=number_of_cuts, mode='list'))
        return self

    @classmethod
    def get_cut_grid(cls, init_window: Union[Tuple[float, float, float, float],
                                             Tuple[float, float]],
                     xy_step: Tuple[float, float],
                     xy_size: Tuple[int, int]):
        if len(init_window) == 2:
            init_window = (init_window[0], init_window[0] + xy_step[0],
                           init_window[1], init_window[1] + xy_step[1])
        self = cls()
        for i in range(xy_size[0]):
            for j in range(xy_size[1]):
                self.append((init_window[0] + xy_step[0] * i,
                             init_window[1] + xy_step[0] * i,
                             init_window[2] + xy_step[1] * j,
                             init_window[3] + xy_step[1] * j))
        setattr(self, 'parameters', dict(init_window=init_window, xy_step=xy_step,
                                         xy_size=xy_size, mode='grid'))
        return self

    @classmethod
    def init_from_parameters_dict(cls, parameters: dict):
        try:
            mode = parameters.pop('mode')
            if mode == 'grid':
                return cls.get_cut_grid(**parameters)
            elif mode == 'list':
                return cls.get_cut_list(**parameters)
            else:
                raise ValueError('Mode should be "grid" or "list".')

        except TypeError as err:
            raise KeyError('Provided parameters dict does not contain'
                           ' necessary parameter. Error: %s ' % err)

    def get_ranges(self) -> dict:
        def func(i):
            return [x[i] for x in self]

        x_range = (min(func(0)), max(func(1)))
        y_range = (min(func(2)), max(func(3)))
        return dict(x_range=x_range, y_range=y_range)

    def __repr__(self):
        return f'<CutList>.\nParameters: {self.parameters}\n' \
            f'Values:{super(CutList, self).__repr__()}>'


class XPCSMap(XPCS):
    DEFAULT_XPCS_MAP_H5_FILEPATH = '/media/vladimir/data/GISAXS_DATA/xpcs2d.h5'

    @staticmethod
    def _default_xpcs_map_h5_group_name(folder_num: int):
        return 'zaptime_%d' % folder_num

    def __init__(self, folder_num: int = None,
                 cut_list: CutList = None,
                 gp: GisaxsParameters = None,
                 h5_filepath: str = None,
                 h5_path: str = None,
                 xpcs_map_h5_filepath: str = None,
                 h5_group_name: str = None):
        super(XPCSMap, self).__init__(folder_num, None,
                                      gp, h5_filepath, h5_path)
        self.cut_list = cut_list
        self.xpcs_map_h5_filepath = xpcs_map_h5_filepath or self.DEFAULT_XPCS_MAP_H5_FILEPATH
        self.group_name = h5_group_name or self._default_xpcs_map_h5_group_name(folder_num)
        self.map_is_saved = False

    def set_cut_grid(self):
        raise NotImplementedError()

    def set_cut_list(self):
        raise NotImplementedError()

    def plot_grid(self):
        raise NotImplementedError()

    def save_map(self, cut_list: CutList = None):
        if cut_list is not None:
            self.cut_list = cut_list
        if self.cut_list is None:
            raise ValueError('cut_list should be provided.')

        def save_dset_to_h5(dset, dset_name, cut):
            with h5py.File(self.xpcs_map_h5_filepath, 'a') as f:
                folder = f[self.group_name]
                dset = folder.create_dataset(dset_name, data=dset)
                dset.attrs['cut'] = cut
                logger.debug(f'XPCS saved to {"/".join([self.xpcs_map_h5_filepath, self.group_name, dset_name])}')

        with h5py.File(self.xpcs_map_h5_filepath, 'a') as f:
            if self.group_name in f.keys():
                del f[self.group_name]
            folder = f.create_group(self.group_name)
            folder.attrs.update(self.cut_list.parameters)

        for i, cut in enumerate(tqdm(self.cut_list)):
            self.set_region(cut)
            res = self.get_two_time_corr_function(mode='std')
            save_dset_to_h5(res, str(i), cut)
            del res

        self.map_is_saved = True

    def plot_grid_on_fly(self, cut_list: CutList = None, figsize: Tuple[int, int] = None,
                         reduce_factor: float = None,
                         reduce_mode: str = 'nearest',
                         save_path: str = None):
        if cut_list is not None:
            self.cut_list = cut_list
        if self.cut_list is None:
            raise ValueError('cut_list should be provided.')

        parameters = self.cut_list.parameters

        if parameters['mode'] != 'grid':
            raise ValueError('Method is allowed only for a grid mode.')

        nx, ny = parameters['xy_size']

        figsize = figsize or (nx, ny)
        fig, axs = plt.subplots(nrows=ny, ncols=nx, sharex=True, figsize=figsize)
        axs = np.flip(axs.transpose(), axis=1).flatten()

        for i, (ax, cut) in enumerate(tqdm(zip(axs, self.cut_list))):
            self.set_region(cut)
            res = self.get_two_time_corr_function(mode='std')
            if reduce_factor:
                res = zoom(res, reduce_factor, mode=reduce_mode)
            ax.imshow(res, cmap='jet')
            ax.set_xticks([], [])
            ax.set_yticks([], [])
            if i < ny:
                ax.set_ylabel(str((cut[2] + cut[3]) / 2)[:5])
            if i % ny == 0:
                ax.set_xlabel(str((cut[0] + cut[1]) / 2)[:5])

        plt.subplots_adjust(wspace=0, hspace=0)
        if not save_path:
            plt.show()
        else:
            plt.savefig(save_path)

    def plot_map(self, figsize: Tuple[int, int] = None,
                 reduce_factor: float = None,
                 reduce_mode: str = 'nearest',
                 save_path: str = None):
        if not self.map_is_saved:
            raise ValueError('Map is not saved. '
                             'Use plot_saved_map method to plot previously saved maps.')
        try:
            mode = self.cut_list.parameters['mode']
        except (KeyError, AttributeError) as err:
            raise ValueError('cut_list attribute is wrong. Error: %s' % err)

        if mode == 'grid':
            self._plot_grid_map(figsize, reduce_factor, reduce_mode)
        elif mode == 'list':
            self._plot_list_map(figsize, reduce_factor, reduce_mode)
        else:
            raise ValueError(f"cut_list.parameters['mode'] should be 'list' or 'grid'. "
                             f"Found {mode} instead.")
        if not save_path:
            plt.show()
        else:
            plt.savefig(save_path)

    def _plot_grid_map(self, figsize: Tuple[int, int] = None,
                       reduce_factor: float = None,
                       reduce_mode: str = 'nearest'):

        with h5py.File(self.xpcs_map_h5_filepath, 'r') as f:
            parameters = dict(f[self.group_name].attrs)

        nx, ny = parameters['xy_size']

        def get_dset(name):
            with h5py.File(self.xpcs_map_h5_filepath, 'r') as f:
                dset = f[self.group_name][name]
                cut = dset.attrs['cut']
                res = dset[()]
                return res, cut

        figsize = figsize or (nx, ny)
        fig, axs = plt.subplots(nrows=ny, ncols=nx, sharex=True, figsize=figsize)
        axs = np.flip(axs.transpose(), axis=1).flatten()

        for i, ax in enumerate(tqdm(axs)):
            res, cut = get_dset(str(i))
            if reduce_factor:
                res = zoom(res, reduce_factor, mode=reduce_mode)
            ax.imshow(res, cmap='jet')
            ax.set_xticks([], [])
            ax.set_yticks([], [])
            if i < ny:
                ax.set_ylabel(str((cut[2] + cut[3]) / 2)[:5])
            if i % ny == 0:
                ax.set_xlabel(str((cut[0] + cut[1]) / 2)[:5])

        plt.subplots_adjust(wspace=0, hspace=0)

    def plot_grid_part(self, xy_size: Tuple[int, int, int, int],
                       figsize: Tuple[int, int] = None,
                       save_path: str = None, reduce_factor: float = None,
                       reduce_mode: str = 'nearest', parameters: dict = None):
        parameters = parameters or self.cut_list.parameters
        nx, ny = parameters['xy_size']

        def get_dset(name):
            with h5py.File(self.xpcs_map_h5_filepath, 'r') as f:
                dset = f[self.group_name][name]
                cut = dset.attrs['cut']
                res = dset[()]
                return res, cut

        figsize = figsize or (nx, ny)
        fig, axs = plt.subplots(nrows=ny, ncols=nx, sharex=True, figsize=figsize)
        axs = np.flip(axs.transpose(), axis=1).flatten()

        for i, ax in enumerate(tqdm(axs)):
            res, cut = get_dset(str(i))
            if reduce_factor:
                res = zoom(res, reduce_factor, mode=reduce_mode)
            ax.imshow(res, cmap='jet')
            ax.set_xticks([], [])
            ax.set_yticks([], [])
            if i < ny:
                ax.set_ylabel(str((cut[2] + cut[3]) / 2)[:5])
            if i % ny == 0:
                ax.set_xlabel(str((cut[0] + cut[1]) / 2)[:5])

        plt.subplots_adjust(wspace=0, hspace=0)

    def _plot_list_map(self, figsize: Tuple[int, int] = None,
                       reduce_factor: float = None,
                       reduce_mode: str = 'nearest', parameters: dict = None):
        # parameters = parameters or cut_list.parameters
        raise NotImplementedError

    def plot_saved_map(self, folder_num: int = None, group_name: str = None,
                       h5_file_path: str = None,
                       save_path: str = None, **kwargs):
        raise NotImplementedError('Need to fix bugs first.')

        if group_name:
            pass
        elif folder_num:
            group_name = self._default_xpcs_map_h5_group_name(folder_num)
        else:
            group_name = self.group_name
        h5_file_path = h5_file_path or self.xpcs_map_h5_filepath

        with h5py.File(h5_file_path, 'r') as f:
            try:
                group = f[group_name]
            except KeyError:
                raise KeyError('Group %s is not in the file %s' % (group_name, h5_file_path))
            parameters = dict(group.attrs)

        mode = parameters.pop('mode')

        if mode == 'grid':
            self._plot_grid_map(parameters=parameters, **kwargs)
        elif mode == 'list':
            self._plot_list_map(parameters=parameters, **kwargs)

        if not save_path:
            plt.show()
        else:
            plt.savefig(save_path)

    def plot_cut(self, filenum: int = 0, **kwargs):
        extract_data = ExtractData()
        extract_data.get_images(self.folder_num, filenum,
                                apply_log=True, plot=True, show=False,
                                gp=self.gp,
                                **self.cut_list.get_ranges())
        ax = plt.gca()
        for cut in self.cut_list:
            r = patches.Rectangle((cut[0], cut[2]), cut[1] - cut[0],
                                  cut[3] - cut[2],
                                  fill=False, color='red', **kwargs)
            ax.add_patch(r)
        plt.show()


def plot_one_time_g(folder_num: int, x_range_cut: Tuple[float, float],
                    delta_x: float, y_position: Tuple[float, float],
                    x_range: Tuple[float, float] = (0, 0.3),
                    y_range: Tuple[float, float] = (0, 0.4)):
    extract_data = ExtractData()
    extract_data.get_images(folder_num, 1, x_range=x_range, y_range=y_range,
                            apply_log=True, plot=True, show=False)
    ax = plt.gca()
    xpcs = XPCS(folder_num)
    f_list = list()
    x_cuts = np.linspace(x_range_cut[0], x_range_cut[1],
                         int((x_range_cut[1] - x_range_cut[0]) / delta_x + 1))
    for i in tqdm(x_cuts):
        cut = (i, i + delta_x, y_position[0], y_position[1])
        r = patches.Rectangle((cut[0], cut[2]), cut[1] - cut[0], cut[3] - cut[2],
                              fill=False, color='red')
        ax.add_patch(r)
        xpcs.set_region(cut)
        f, t = xpcs.get_one_time_corr_function()
        f_list.append(f)

    plt.show()
    plt.figure()
    for i, f in enumerate(f_list):
        plt.plot(t, f, label=str(x_cuts[i])[:4])
    plt.legend()
    plt.show()
    return f_list, t


def show_copied_folders():
    xpcs = XPCS(96)
    print(xpcs.DEFAULT_H5_FILEPATH)
    with h5py.File(xpcs.DEFAULT_H5_FILEPATH, 'r') as f:
        pprint(list(f['Raw data'].keys()))


def save_xpcs_maps(folder_list: List[int], cut_list: CutList = None, *,
                   reduce_factor: float = 0.2):
    cut_list = cut_list or CutList.get_cut_grid((-0.1, 0),
                                                xy_size=(10, 15),
                                                xy_step=(0.02, 0.02))
    for folder_number in tqdm(folder_list):
        xpcs_map = XPCSMap(folder_number, cut_list=cut_list)
        xpcs_map.save_map()
        xpcs_map.plot_map(reduce_factor=reduce_factor,
                          save_path=f'xpcs_map_{folder_number}.png')


def xpcs_2d_profiler():
    import cProfile
    cProfile.run('XPCS(20, (0.03, 0.04, 0.09, 0.1)).get_two_time_corr_function()')


if __name__ == '__main__':
    logging.basicConfig(level=logging.ERROR)
    # show_copied_folders()
    folder_number = 38
    cut_list = CutList.get_cut_grid((-0.1, 0),
                                    xy_size=(10, 15),
                                    xy_step=(0.02, 0.02))
    xpcs_map = XPCSMap(folder_number, cut_list=cut_list)
    xpcs_map.plot_saved_map(folder_number, reduce_factor=0.4,
                            save_path=f'xpcs_map_{folder_number}.png')
    #
    # for folder in [69, 76, 85, 99]:
    #     try:
    #         xpcs_map = XPCSMap(folder_number, cut_list=cut_list)
    #         # xpcs_map.save_map()
    #         xpcs_map.plot_saved_map(folder_number, reduce_factor=0.2)
    #     except Exception as err:
    #         logger.error(err)
    # xpcs_map.plot_saved_map(folder_number, reduce_factor=0.05)

    # xpcs = XPCS(96, (0.03, 0.04, 0.09, 0.1))
    # xpcs_2d_profiler(xpcs)
    # [20, 38, 69, 76, 85, 90, 99]



# -*- coding: utf-8 -*-

from collections import defaultdict
from typing import Tuple, List
import os
from pprint import pprint
import logging

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib as mpl
from matplotlib import cm
from matplotlib import rc
from matplotlib.colors import LogNorm

import numpy as np
from scipy.ndimage.filters import gaussian_filter1d
from scipy.optimize import curve_fit
from scipy.fftpack import fft

import h5py
from tqdm import tqdm

from .extract_data import ExtractData, get_time_axis, get_temp_axis
from .common_tools import save_create, get_zaptime_files, get_pen_folder_number
from .gisaxs_parameters import GisaxsParameters
from .xpcs import XPCS, plot_one_time_g

logging.basicConfig(level=logging.ERROR,
                    format='%(asctime)s %(levelname)s:%(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p')

ed0 = ed1 = ExtractData()

labels_size = 20
mpl.rcParams['text.latex.preamble'] = [r'\boldmath']
rc('text', usetex=True)
rc('xtick', labelsize=labels_size)
rc('ytick', labelsize=labels_size)
rc('axes', labelsize=labels_size)
rc('axes', titlesize=labels_size)
rc('axes', titlesize=labels_size)
rc('legend', fontsize=labels_size)



from setuptools import setup, find_packages

setup(name='GISAXS_XPCS',
      version='0.1.1',
      description='Package for GISAXS-XPCS analysis. For internal use within Shreiber lab.',
      author='Vladimir Starostin',
      author_email='vladimir.starostin@uni-tuebingen.de',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'numpy', 'pandas', 'matplotlib', 'scipy', 'h5py', 'tqdm', 'jupyter'
      ],
      include_package_data=True,
      zip_safe=False)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple

import numpy as np


class GisaxsParameters(object):
    def __init__(self, *,
                 image_size: Tuple[int, int] = (1030, 514),
                 beam_center: Tuple[int, int] = (314, 265),
                 wavelength: float = 1.5307,
                 angle_of_incidence_deg: float = 0.1,
                 detector_distance: int = 5292,
                 pixel_size: float = 0.075):
        """

        :param image_size: image size in pixels
        :param beam_center: tuple of beam center pixels coordinates
        :param wavelength: wavelength [A]
        :param angle_of_incidence_deg: [deg]
        :param detector_distance: [mm]
        :param pixel_size: [mm]
        """
        self.size = image_size
        self.beam_center = beam_center
        self.wavelength = wavelength
        self.angle_of_incidence_deg = angle_of_incidence_deg
        self.angle_of_incidence = angle_of_incidence_deg / 180 * np.pi
        self.detector_distance = detector_distance
        self.pixel_size = pixel_size
        self.k0 = 2 * np.pi / self.wavelength

    @classmethod
    def init_from_folder_number(cls, folder_num: int):
        if folder_num <= 42:
            return cls(beam_center=(247, 265))
        else:
            return cls()

    def get_angle_vectors(self, *,
                          flip: bool = True, units: str = 'rad') -> Tuple[np.array, np.array]:
        """
        Returns image axes in provided units (degrees or radians).

        :param flip: if True, flips both axes (True by default)
        :param units: 'rad' or 'deg'
        :return: two numpy arrays of image axes
        """
        if units not in ('rad', 'deg'):
            raise ValueError('Unknown units %s, should be deg or rad' % units)
        x_vector = np.flip(np.linspace(1, self.size[1], self.size[1]) - self.size[1] +
                           self.beam_center[0], axis=0) * self.pixel_size
        y_vector = np.flip(np.linspace(1, self.size[0], self.size[0]) -
                           self.beam_center[1], axis=0) * self.pixel_size - \
                   self.detector_distance * np.tan(
            self.angle_of_incidence)
        angle_x = np.arctan(x_vector / self.detector_distance)
        angle_y = np.arctan(y_vector / self.detector_distance)
        if flip:
            angle_x = np.flip(angle_x, axis=0)
            angle_y = np.flip(angle_y, axis=0)
        if units == 'deg':
            angle_x *= 180 / np.pi
            angle_y *= 180 / np.pi
        return angle_x, angle_y

    def get_q_vectors(self, back_reshape: bool = True):
        angle_xx, angle_yy = self.get_angle_vectors()
        size = len(angle_xx) * len(angle_yy)
        angle_x, angle_y = np.meshgrid(angle_xx, angle_yy)
        angle_x = angle_x.reshape((1, size))[0]
        angle_y = angle_y.reshape((1, size))[0]
        q_xy = self.k0 * np.sign(angle_x) * \
               np.sqrt((np.cos(angle_y) * np.cos(angle_x) - np.cos(self.angle_of_incidence)) ** 2 +
                       (np.cos(angle_y) * np.sin(angle_x)) ** 2)
        q_z = self.k0 * (np.sin(angle_y) + np.sin(self.angle_of_incidence))

        if back_reshape:
            q_xy = self.back_reshape(q_xy)
            q_z = self.back_reshape(q_z)
        return [q_xy, q_z]

    def back_reshape(self, data):
        data = np.reshape(data, (self.size[0], self.size[1]))
        data = np.flip(data, axis=0)
        data = np.rot90(data, k=1)
        data = np.flip(data, axis=1)
        data = np.reshape(data, (1, self.size[1] * self.size[0]))[0]
        return data
